@@ -19,15 +19,14 @@ import tensorflow_probability as tfp
 from typing_extensions import final
 
 from ..data import Dataset
-from ..type import QueryPoints
 from ..models import ProbabilisticModel
+from ..type import TensorType
+from ..utils import DEFAULTS
 
-
-AcquisitionFunction = Callable[[QueryPoints], tf.Tensor]
+AcquisitionFunction = Callable[[TensorType], TensorType]
 """
 Type alias for acquisition functions.
 
-
 An `AcquisitionFunction` maps a single query point (of dimension `D`) to a single value that
 describes how useful it would be evaluate that point (to our goal of optimizing the objective
 function). Thus, with leading dimensions, an `AcquisitionFunction` takes input shape
@@ -37,9 +36,9 @@ function). Thus, with leading dimensions, an `AcquisitionFunction` takes input s
 `BatchAcquisitionFunction`.
 """
 
-BatchAcquisitionFunction = Callable[[QueryPoints], tf.Tensor]
-""" 
-Type alias for batch acquisition functions. 
+BatchAcquisitionFunction = Callable[[TensorType], TensorType]
+"""
+Type alias for batch acquisition functions.
 
 A `BatchAcquisitionFunction` maps a set of `B` query points (each of dimension `D`) to a single
 value that describes how useful it would be evaluate all these points together (to our goal of
@@ -107,6 +106,10 @@ class ExpectedImprovement(SingleModelAcquisitionBuilder):
     of the posterior mean at observed points.
     """
 
+    def __repr__(self) -> str:
+        """"""
+        return "ExpectedImprovement()"
+
     def prepare_acquisition_function(
         self, dataset: Dataset, model: ProbabilisticModel
     ) -> AcquisitionFunction:
@@ -125,12 +128,12 @@ class ExpectedImprovement(SingleModelAcquisitionBuilder):
 
     @staticmethod
     def _acquisition_function(
-        model: ProbabilisticModel, eta: tf.Tensor, at: QueryPoints
-    ) -> tf.Tensor:
+        model: ProbabilisticModel, eta: TensorType, at: TensorType
+    ) -> TensorType:
         return expected_improvement(model, eta, at)
 
 
-def expected_improvement(model: ProbabilisticModel, eta: tf.Tensor, at: QueryPoints) -> tf.Tensor:
+def expected_improvement(model: ProbabilisticModel, eta: TensorType, at: TensorType) -> TensorType:
     r"""
     The Expected Improvement (EI) acquisition function for single-objective global optimization.
     Return the expectation of the improvement at ``at`` over the current "best" observation ``eta``,
@@ -139,20 +142,7 @@ def expected_improvement(model: ProbabilisticModel, eta: tf.Tensor, at: QueryPoi
 
     .. math:: x \mapsto \mathbb E \left[ \max (\eta - f(x), 0) \right]
 
-    This function was introduced by Mockus et al, 1975. See the following for details:
-
-    ::
-
-       @article{Jones:1998,
-            title={Efficient global optimization of expensive black-box functions},
-            author={Jones, Donald R and Schonlau, Matthias and Welch, William J},
-            journal={Journal of Global optimization},
-            volume={13},
-            number={4},
-            pages={455--492},
-            year={1998},
-            publisher={Springer}
-       }
+    This function was introduced by Mockus et al, 1975. See :cite:`Jones:1998` for details.
 
     :param model: The model of the objective function.
     :param eta: The "best" observation.
@@ -177,6 +167,10 @@ class NegativeLowerConfidenceBound(SingleModelAcquisitionBuilder):
         """
         self._beta = beta
 
+    def __repr__(self) -> str:
+        """"""
+        return f"NegativeLowerConfidenceBound({self._beta!r})"
+
     def prepare_acquisition_function(
         self, dataset: Dataset, model: ProbabilisticModel
     ) -> AcquisitionFunction:
@@ -189,7 +183,7 @@ class NegativeLowerConfidenceBound(SingleModelAcquisitionBuilder):
         return lambda at: self._acquisition_function(model, self._beta, at)
 
     @staticmethod
-    def _acquisition_function(model: ProbabilisticModel, beta: float, at: QueryPoints) -> tf.Tensor:
+    def _acquisition_function(model: ProbabilisticModel, beta: float, at: TensorType) -> TensorType:
         return -lower_confidence_bound(model, beta, at)
 
 
@@ -202,26 +196,18 @@ class NegativePredictiveMean(NegativeLowerConfidenceBound):
     def __init__(self):
         super().__init__(beta=0.0)
 
+    def __repr__(self) -> str:
+        """"""
+        return "NegativePredictiveMean()"
 
-def lower_confidence_bound(model: ProbabilisticModel, beta: float, at: QueryPoints) -> tf.Tensor:
+
+def lower_confidence_bound(model: ProbabilisticModel, beta: float, at: TensorType) -> TensorType:
     r"""
     The lower confidence bound (LCB) acquisition function for single-objective global optimization.
 
     .. math:: x^* \mapsto \mathbb{E} [f(x^*)|x, y] - \beta \sqrt{ \mathrm{Var}[f(x^*)|x, y] }
 
-    See the following for details:
-
-    ::
-
-        @inproceedings{Srinivas:2010,
-            author = "Srinivas, Niranjan and Krause, Andreas and Seeger, Matthias and Kakade, Sham M.",
-            booktitle = "{Proceedings of the 27th International Conference on Machine Learning (ICML-10)}",
-            editor = "F{\"u}rnkranz, Johannes and Joachims, Thorsten",
-            pages = "1015--1022",
-            publisher = "Omnipress",
-            title = "{Gaussian Process Optimization in the Bandit Setting: No Regret and Experimental Design}",
-            year = "2010"
-        }
+    See :cite:`Srinivas:2010` for details.
 
     :param model: The model of the objective function.
     :param beta: The weight to give to the standard deviation contribution of the LCB. Must not be
@@ -242,40 +228,18 @@ def lower_confidence_bound(model: ProbabilisticModel, beta: float, at: QueryPoin
 class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder):
     r"""
     Builder for the :func:`probability_of_feasibility` acquisition function, defined in
-    Garner, 2014 as
+    :cite:`gardner14` as
 
-    .. math:: \int_{-\infty}^{\tau} p(c(\mathbf{x}) | \mathbf{x}, \mathcal{D}) \mathrm{d} c(\mathbf{x}) \qquad ,
+    .. math::
+
+        \int_{-\infty}^{\tau} p(c(\mathbf{x}) | \mathbf{x}, \mathcal{D}) \mathrm{d} c(\mathbf{x})
+        \qquad ,
 
     where :math:`\tau` is a threshold. Values below the threshold are considered feasible by the
-    constraint function. See the following for details:
-
-    ::
-
-        @inproceedings{gardner14,
-            title={Bayesian Optimization with Inequality Constraints},
-            author={Jacob Gardner and Matt Kusner and Zhixiang and Kilian Weinberger and John Cunningham},
-            booktitle={Proceedings of the 31st International Conference on Machine Learning},
-            year={2014},
-            volume={32},
-            number={2},
-            series={Proceedings of Machine Learning Research},
-            month={22--24 Jun},
-            publisher={PMLR},
-            url={http://proceedings.mlr.press/v32/gardner14.html},
-        }
-
-        @article{schonlau1998global,
-            title={Global versus local search in constrained optimization of computer models},
-            author={Schonlau, Matthias and Welch, William J and Jones, Donald R},
-            journal={Lecture Notes-Monograph Series},
-            pages={11--25},
-            year={1998},
-            publisher={JSTOR}
-        }
-
+    constraint function. See also :cite:`schonlau1998global` for details.
     """
 
-    def __init__(self, threshold: Union[float, tf.Tensor]):
+    def __init__(self, threshold: Union[float, TensorType]):
         """
         :param threshold: The (scalar) probability of feasibility threshold.
         :raise ValueError (or InvalidArgumentError): If ``threshold`` is not a scalar.
@@ -286,8 +250,12 @@ class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder):
 
         self._threshold = threshold
 
+    def __repr__(self) -> str:
+        """"""
+        return f"ProbabilityOfFeasibility({self._threshold!r})"
+
     @property
-    def threshold(self) -> Union[float, tf.Tensor]:
+    def threshold(self) -> Union[float, TensorType]:
         """ The probability of feasibility threshold. """
         return self._threshold
 
@@ -303,18 +271,21 @@ class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder):
 
     @staticmethod
     def _acquisition_function(
-        model: ProbabilisticModel, threshold: Union[float, tf.Tensor], at: QueryPoints
-    ) -> tf.Tensor:
+        model: ProbabilisticModel, threshold: Union[float, TensorType], at: TensorType
+    ) -> TensorType:
         return probability_of_feasibility(model, threshold, at)
 
 
 def probability_of_feasibility(
-    model: ProbabilisticModel, threshold: Union[float, tf.Tensor], at: QueryPoints
-) -> tf.Tensor:
+    model: ProbabilisticModel, threshold: Union[float, TensorType], at: TensorType
+) -> TensorType:
     r"""
-    The probability of feasibility acquisition function defined in Garner, 2014 as
+    The probability of feasibility acquisition function defined in :cite:`gardner14` as
 
-    .. math:: \int_{-\infty}^{\tau} p(c(\mathbf{x}) | \mathbf{x}, \mathcal{D}) \mathrm{d}c(\mathbf{x}) \qquad ,
+    .. math::
+
+        \int_{-\infty}^{\tau} p(c(\mathbf{x}) | \mathbf{x}, \mathcal{D}) \mathrm{d} c(\mathbf{x})
+        \qquad ,
 
     where :math:`\tau` is a threshold. Values below the threshold are considered feasible by the
     constraint function.
@@ -336,35 +307,17 @@ def probability_of_feasibility(
 
 class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder):
     """
-    Builder for the _expected constrained improvement_ acquisition function defined in
-    Gardner, 2014. The acquisition function computes the expected improvement from the best
+    Builder for the *expected constrained improvement* acquisition function defined in
+    :cite:`gardner14`. The acquisition function computes the expected improvement from the best
     feasible point, where feasible points are those that (probably) satisfy some constraint. Where
     there are no feasible points, this builder simply builds the constraint function.
-
-    See the following for details:
-
-    ::
-
-        @inproceedings{gardner14,
-            title={Bayesian Optimization with Inequality Constraints},
-            author={Jacob Gardner and Matt Kusner and Zhixiang and Kilian Weinberger and John Cunningham},
-            booktitle={Proceedings of the 31st International Conference on Machine Learning},
-            year={2014},
-            volume={32},
-            number={2},
-            series={Proceedings of Machine Learning Research},
-            month={22--24 Jun},
-            publisher={PMLR},
-            url={http://proceedings.mlr.press/v32/gardner14.html},
-        }
-
     """
 
     def __init__(
         self,
         objective_tag: str,
         constraint_builder: AcquisitionFunctionBuilder,
-        min_feasibility_probability: Union[float, tf.Tensor] = 0.5,
+        min_feasibility_probability: Union[float, TensorType] = 0.5,
     ):
         """
         :param objective_tag: The tag for the objective data and model.
@@ -386,6 +339,13 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder):
         self._constraint_builder = constraint_builder
         self._min_feasibility_probability = min_feasibility_probability
 
+    def __repr__(self) -> str:
+        """"""
+        return (
+            f"ExpectedConstrainedImprovement({self._objective_tag!r}, {self._constraint_builder!r},"
+            f" {self._min_feasibility_probability!r})"
+        )
+
     def prepare_acquisition_function(
         self, datasets: Mapping[str, Dataset], models: Mapping[str, ProbabilisticModel]
     ) -> AcquisitionFunction:
@@ -438,12 +398,19 @@ class IndependentReparametrizationSampler:
         tf.debugging.assert_positive(sample_size)
 
         self._sample_size = sample_size
+
+        # _eps is essentially a lazy constant. It is declared and assigned an empty tensor here, and
+        # populated on the first call to sample
         self._eps = tf.Variable(
             tf.ones([sample_size, 0], dtype=tf.float64), shape=[sample_size, None]
         )  # [S, 0]
         self._model = model
 
-    def sample(self, at: QueryPoints) -> tf.Tensor:
+    def __repr__(self) -> str:
+        """"""
+        return f"IndependentReparametrizationSampler({self._sample_size!r}, {self._model!r})"
+
+    def sample(self, at: TensorType) -> TensorType:
         """
         Return approximate samples from the `model` specified at :meth:`__init__`. Multiple calls to
         :meth:`sample`, for any given :class:`IndependentReparametrizationSampler` and ``at``, will
@@ -457,14 +424,14 @@ class IndependentReparametrizationSampler:
         :raise ValueError (or InvalidArgumentError): If ``at`` is a scalar.
         """
         tf.debugging.assert_rank_at_least(at, 1)
-        mean, cov = self._model.predict(at[..., None, :])  # [..., 1, L], [..., 1, L]
+        mean, var = self._model.predict(at[..., None, :])  # [..., 1, L], [..., 1, L]
 
         if tf.size(self._eps) == 0:
             self._eps.assign(
                 tf.random.normal([self._sample_size, mean.shape[-1]], dtype=tf.float64)
             )  # [S, L]
 
-        return mean + tf.sqrt(cov) * tf.cast(self._eps, cov.dtype)  # [..., S, L]
+        return mean + tf.sqrt(var) * tf.cast(self._eps, var.dtype)  # [..., S, L]
 
 
 class MCIndAcquisitionFunctionBuilder(AcquisitionFunctionBuilder):
@@ -473,6 +440,10 @@ class MCIndAcquisitionFunctionBuilder(AcquisitionFunctionBuilder):
     estimates the value of evaluating the observer at a given point, and does this using Monte-Carlo
     estimation via the reparameterization trick. This class is essentially a convenience
     :class:`AcquisitionFunctionBuilder` using a :class:`IndependentReparametrizationSampler`.
+
+    Subclasses implement :meth:`_build_with_sampler` which, in addition to the arguments `datasets`
+    and `models`, provides a :class:`IndependentReparametrizationSampler` for each model which can
+    be used to approximate continuous samples from the models.
     """
 
     def __init__(self, sample_size: int):
@@ -513,9 +484,9 @@ class SingleModelMCIndAcquisitionFunctionBuilder(SingleModelAcquisitionBuilder):
     A :class:`SingleModelMCIndAcquisitionFunctionBuilder` builds an acquisition function that
     estimates the value of evaluating the observer at a given point, and does this using Monte-Carlo
     estimation via the reparameterization trick. This class is essentially a convenience
-    :class:`AcquisitionFunctionBuilder` using a :class:`IndependentReparametrizationSampler`.
+    :class:`SingleModelAcquisitionBuilder` using a :class:`IndependentReparametrizationSampler`.
 
-    Subclasses implement :meth:`_build_with_samples` which, in addition to the arguments `dataset`
+    Subclasses implement :meth:`_build_with_sampler` which, in addition to the arguments `dataset`
     and `model`, provides a :class:`IndependentReparametrizationSampler` which can be used to
     approximate continuous samples from the model.
     """
@@ -551,7 +522,10 @@ class SingleModelMCIndAcquisitionFunctionBuilder(SingleModelAcquisitionBuilder):
 
 
 class BatchAcquisitionFunctionBuilder(ABC):
-    """ A :class:`BatchAcquisitionFunctionBuilder` builds a batch acquisition function. """
+    """
+    A :class:`BatchAcquisitionFunctionBuilder` builds an acquisition function for evaluating batches
+    of query points.
+    """
 
     @abstractmethod
     def prepare_acquisition_function(
@@ -562,3 +536,194 @@ class BatchAcquisitionFunctionBuilder(ABC):
         :param models: The models over each dataset in ``datasets``.
         :return: A batch acquisition function.
         """
+
+
+class SingleModelBatchAcquisitionBuilder(ABC):
+    """
+    Convenience acquisition function builder for a batch acquisition function (or component of a
+    composite batch acquisition function) that requires only one model, dataset pair.
+    """
+
+    def using(self, tag: str) -> BatchAcquisitionFunctionBuilder:
+        """
+        :param tag: The tag for the model, dataset pair to use to build this acquisition function.
+        :return: A batch acquisition function builder that selects the model and dataset specified
+            by ``tag``, as defined in :meth:`prepare_acquisition_function`.
+        """
+        single_builder = self
+
+        class _Anon(BatchAcquisitionFunctionBuilder):
+            def prepare_acquisition_function(
+                self, datasets: Mapping[str, Dataset], models: Mapping[str, ProbabilisticModel]
+            ) -> AcquisitionFunction:
+                return single_builder.prepare_acquisition_function(datasets[tag], models[tag])
+
+            def __repr__(self) -> str:
+                return f"{single_builder!r} using tag {tag!r}"
+
+        return _Anon()
+
+    @abstractmethod
+    def prepare_acquisition_function(
+        self, dataset: Dataset, model: ProbabilisticModel
+    ) -> AcquisitionFunction:
+        """
+        :param dataset: The data to use to build the acquisition function.
+        :param model: The model over the specified ``dataset``.
+        :return: A batch acquisition function.
+        """
+
+
+class BatchReparametrizationSampler:
+    r"""
+    This sampler employs the *reparameterization trick* to approximate batches of samples from a
+    :class:`ProbabilisticModel`\ 's predictive joint distribution as
+
+    .. math:: x \mapsto \mu(x) + \epsilon L(x)
+
+    where :math:`L` is the Cholesky factor s.t. :math:`LL^T` is the covariance, and
+    :math:`\epsilon \sim \mathcal N (0, 1)` is constant for a given sampler, thus ensuring samples
+    form a continuous curve.
+    """
+
+    def __init__(self, sample_size: int, model: ProbabilisticModel):
+        """
+        :param sample_size: The number of samples for each batch of points. Must be positive.
+        :param model: The model to sample from.
+        :raise ValueError (or InvalidArgumentError): If ``sample_size`` is not positive.
+        """
+        tf.debugging.assert_positive(sample_size)
+
+        self._sample_size = sample_size
+
+        # _eps is essentially a lazy constant. It is declared and assigned an empty tensor here, and
+        # populated on the first call to sample
+        self._eps = tf.Variable(
+            tf.ones([0, 0, sample_size], dtype=tf.float64), shape=[None, None, sample_size]
+        )  # [0, 0, S]
+        self._model = model
+
+    def __repr__(self) -> str:
+        """"""
+        return f"BatchReparametrizationSampler({self._sample_size!r}, {self._model!r})"
+
+    def sample(self, at: TensorType, *, jitter: float = DEFAULTS.JITTER) -> TensorType:
+        """
+        Return approximate samples from the `model` specified at :meth:`__init__`. Multiple calls to
+        :meth:`sample`, for any given :class:`BatchReparametrizationSampler` and ``at``, will
+        produce the exact same samples. Calls to :meth:`sample` on *different*
+        :class:`BatchReparametrizationSampler` instances will produce different samples.
+
+        :param at: Batches of query points at which to sample the predictive distribution, with
+            shape `[..., B, D]`, for batches of size `B` of points of dimension `D`. Must have a
+            consistent batch size across all calls to :meth:`sample` for any given
+            :class:`BatchReparametrizationSampler`.
+        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
+            the covariance matrix.
+        :return: The samples, of shape `[..., S, B, L]`, where `S` is the `sample_size`, `B` the
+            number of points per batch, and `L` the dimension of the model's predictive
+            distribution.
+        :raise ValueError (or InvalidArgumentError): If any of the following are true:
+
+            - ``at`` is a scalar.
+            - The batch size `B` of ``at`` is not positive.
+            - The batch size `B` of ``at`` differs from that of previous calls.
+            - ``jitter`` is negative.
+        """
+        tf.debugging.assert_rank_at_least(at, 2)
+        tf.debugging.assert_greater_equal(jitter, 0.0)
+
+        batch_size = at.shape[-2]
+
+        tf.debugging.assert_positive(batch_size)
+
+        eps_is_populated = tf.size(self._eps) != 0
+
+        if eps_is_populated:
+            tf.debugging.assert_equal(
+                batch_size,
+                tf.shape(self._eps)[-2],
+                f"{type(self).__name__} requires a fixed batch size. Got batch size {batch_size}"
+                f" but previous batch size was {tf.shape(self._eps)[-2]}.",
+            )
+
+        mean, cov = self._model.predict_joint(at)  # [..., B, L], [..., L, B, B]
+
+        if not eps_is_populated:
+            self._eps.assign(
+                tf.random.normal(
+                    [mean.shape[-1], batch_size, self._sample_size], dtype=tf.float64
+                )  # [L, B, S]
+            )
+
+        identity = tf.eye(batch_size, dtype=cov.dtype)  # [B, B]
+        cov_cholesky = tf.linalg.cholesky(cov + jitter * identity)  # [..., L, B, B]
+
+        variance_contribution = cov_cholesky @ tf.cast(self._eps, cov.dtype)  # [..., L, B, S]
+
+        leading_indices = tf.range(tf.rank(variance_contribution) - 3)
+        absolute_trailing_indices = [-1, -2, -3] + tf.rank(variance_contribution)
+        new_order = tf.concat([leading_indices, absolute_trailing_indices], axis=0)
+
+        return mean[..., None, :, :] + tf.transpose(variance_contribution, new_order)
+
+
+class BatchMonteCarloExpectedImprovement(SingleModelBatchAcquisitionBuilder):
+    """
+    Expected improvement for batches of points (or :math:`q`-EI), approximated using Monte Carlo
+    estimation with the reparametrization trick. See :cite:`Ginsbourger2010` for details.
+
+    Improvement is measured with respect to the minimum predictive mean at observed query points.
+    This is calculated in :class:`BatchMonteCarloExpectedImprovement` by assuming observations
+    at new points are independent from those at known query points. This is faster, but is an
+    approximation for noisy observers.
+    """
+
+    def __init__(self, sample_size: int, *, jitter: float = DEFAULTS.JITTER):
+        """
+        :param sample_size: The number of samples for each batch of points.
+        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
+            the covariance matrix.
+        :raise ValueError (or InvalidArgumentError): If ``sample_size`` is not positive, or
+            ``jitter`` is negative.
+        """
+        tf.debugging.assert_positive(sample_size)
+        tf.debugging.assert_greater_equal(jitter, 0.0)
+
+        super().__init__()
+
+        self._sample_size = sample_size
+        self._jitter = jitter
+
+    def __repr__(self) -> str:
+        """"""
+        return f"BatchMonteCarloExpectedImprovement({self._sample_size!r}, jitter={self._jitter!r})"
+
+    def prepare_acquisition_function(
+        self, dataset: Dataset, model: ProbabilisticModel
+    ) -> AcquisitionFunction:
+        """
+        :param dataset: The data from the observer. Must be populated.
+        :param model: The model over the specified ``dataset``. Must have event shape [1].
+        :return: The batch *expected improvement* acquisition function.
+        :raise ValueError (or InvalidArgumentError): If ``dataset`` is not populated, or ``model``
+            does not have an event shape of [1].
+        """
+        tf.debugging.assert_positive(len(dataset))
+
+        mean, _ = model.predict(dataset.query_points)
+
+        tf.debugging.assert_shapes(
+            [(mean, ["_", 1])], message="Expected model with event shape [1]."
+        )
+
+        eta = tf.reduce_min(mean, axis=0)
+        sampler = BatchReparametrizationSampler(self._sample_size, model)
+
+        def batch_ei(at: TensorType) -> TensorType:
+            samples = tf.squeeze(sampler.sample(at, jitter=self._jitter), axis=-1)  # [..., S, B]
+            min_sample_per_batch = tf.reduce_min(samples, axis=-1)  # [..., S]
+            batch_improvement = tf.maximum(eta - min_sample_per_batch, 0.0)  # [..., S]
+            return tf.reduce_mean(batch_improvement, axis=-1, keepdims=True)  # [..., 1]
+
+        return batch_ei
