@@ -13,8 +13,9 @@
 # limitations under the License.
 """ This module contains implementations of various types of search space. """
 from __future__ import annotations
-from abc import abstractmethod, ABC
-from typing import Union, Type, TypeVar
+
+from abc import ABC, abstractmethod
+from typing import Dict, List, TypeVar, Union, overload
 
 import tensorflow as tf
 
@@ -22,6 +23,7 @@ from .type import TensorType
 from .utils import shapes_equal
 
 SP = TypeVar("SP", bound="SearchSpace")
+""" A type variable bound to :class:`SearchSpace`. """
 
 
 class SearchSpace(ABC):
@@ -30,18 +32,18 @@ class SearchSpace(ABC):
     """
 
     @abstractmethod
-    def sample(self, num_samples: int) -> tf.Tensor:
+    def sample(self, num_samples: int) -> TensorType:
         """
         :param num_samples: The number of points to sample from this search space.
         :return: ``num_samples`` i.i.d. random points, sampled uniformly from this search space.
         """
 
     @abstractmethod
-    def __contains__(self, value: TensorType) -> Union[bool, tf.Tensor]:
+    def __contains__(self, value: TensorType) -> Union[bool, TensorType]:
         """
         :param value: A point to check for membership of this :class:`SearchSpace`.
         :return: `True` if ``value`` is a member of this search space, else `False`. May return a
-            scalar boolean `tf.Tensor` instead of the `bool` itself.
+            scalar boolean `TensorType` instead of the `bool` itself.
         :raise ValueError (or InvalidArgumentError): If ``value`` has a different dimensionality
             from this :class:`SearchSpace`.
         """
@@ -50,23 +52,28 @@ class SearchSpace(ABC):
     def __mul__(self: SP, other: SP) -> SP:
         """
         :param other: A search space of the same type as this search space.
-        :return: The Cartesian product of this search space with ``other``.
+        :return: The Cartesian product of this search space with the ``other``.
         """
 
     def __pow__(self: SP, other: int) -> SP:
         """
-        Return the Cartesian product of ``other`` instances of this search space. For example, for an exponent of `3`, and search space `s`, this is `s ** 3`, which is equivalent to `s * s * s`.
+        Return the Cartesian product of ``other`` instances of this search space. For example, for
+        an exponent of `3`, and search space `s`, this is `s ** 3`, which is equivalent to
+        `s * s * s`.
 
-        :param other: The number of instances of this search space to multiply. Must be strictly positive.
+        :param other: The exponent, or number of instances of this search space to multiply
+            together. Must be strictly positive.
         :return: The Cartesian product of ``other`` instances of this search space.
         :raise ValueError: If the exponent ``other`` is less than 1.
         """
         if other < 1:
-            raise ValueError("The exponent ``other`` can only be a strictly positive integer")
+            raise ValueError(f"Exponent must be strictly positive, got {other}")
 
         space = self
+
         for _ in range(other - 1):
             space *= self
+
         return space
 
 
@@ -92,16 +99,20 @@ class DiscreteSearchSpace(SearchSpace):
         tf.debugging.assert_shapes([(points, ("N", "D"))])
         self._points = points
 
+    def __repr__(self) -> str:
+        """"""
+        return f"DiscreteSearchSpace({self._points!r})"
+
     @property
     def points(self) -> TensorType:
         """ All the points in this space. """
         return self._points
 
-    def __contains__(self, value: TensorType) -> Union[bool, tf.Tensor]:
+    def __contains__(self, value: TensorType) -> Union[bool, TensorType]:
         tf.debugging.assert_shapes([(value, self.points.shape[1:])])
         return tf.reduce_any(tf.reduce_all(value == self._points, axis=1))
 
-    def sample(self, num_samples: int) -> tf.Tensor:
+    def sample(self, num_samples: int) -> TensorType:
         """
         :param num_samples: The number of points to sample from this search space.
         :return: ``num_samples`` i.i.d. random points, sampled uniformly, and without replacement,
@@ -117,7 +128,7 @@ class DiscreteSearchSpace(SearchSpace):
         return tf.random.shuffle(self._points)[:num_samples, :]
 
     def __mul__(self, other: DiscreteSearchSpace) -> DiscreteSearchSpace:
-        """
+        r"""
         Return the Cartesian product of the two :class:`DiscreteSearchSpace`\ s. For example:
 
             >>> sa = DiscreteSearchSpace(tf.constant([[0, 1], [2, 3]]))
@@ -128,20 +139,23 @@ class DiscreteSearchSpace(SearchSpace):
                    [2, 3, 4, 5, 6],
                    [2, 3, 7, 8, 9]], dtype=int32)
 
-        :param other: :class:`DiscreteSearchSpace`.
-        :return: the new combined :class:`DiscreteSearchSpace`
-        :raise TypeError: If the lhs and rhs :class:`DiscreteSearchSpace` points have different types.
+        :param other: A :class:`DiscreteSearchSpace` with :attr:`points` of the same dtype as this
+            search space.
+        :return: The Cartesian product of the two :class:`DiscreteSearchSpace`\ s.
+        :raise TypeError: If one :class:`DiscreteSearchSpace` has :attr:`points` of a different
+            dtype to the other.
         """
-        if self._points.dtype is not other._points.dtype:
+        if self.points.dtype is not other.points.dtype:
             return NotImplemented
 
-        N = self._points.shape[0]
-        M = other._points.shape[0]
-        tile_self = tf.tile(tf.expand_dims(self.points, 1), [1, M, 1])
-        tile_dss = tf.tile(tf.expand_dims(other.points, 0), [N, 1, 1])
-        cartesian_product = tf.concat([tile_self, tile_dss], axis=2)
+        tile_self = tf.tile(self.points[:, None], [1, len(other.points), 1])
+        tile_other = tf.tile(other.points[None], [len(self.points), 1, 1])
+        cartesian_product = tf.concat([tile_self, tile_other], axis=2)
+        product_space_dimension = self.points.shape[-1] + other.points.shape[-1]
+        return DiscreteSearchSpace(tf.reshape(cartesian_product, [-1, product_space_dimension]))
 
-        return DiscreteSearchSpace(tf.reshape(cartesian_product, [N * M, -1]))
+    def __deepcopy__(self, memo: Dict[int, object]) -> DiscreteSearchSpace:
+        return self
 
 
 class Box(SearchSpace):
@@ -151,35 +165,52 @@ class Box(SearchSpace):
     closed bounded intervals in :math:`\mathbb{R}`.
     """
 
+    @overload
+    def __init__(self, lower: List[float], upper: List[float]):
+        ...
+
+    @overload
     def __init__(self, lower: TensorType, upper: TensorType):
+        ...
+
+    def __init__(
+        self, lower: Union[List[float], TensorType], upper: Union[List[float], TensorType]
+    ):
+        r"""
+        If ``lower`` and ``upper`` are `list`\ s, they will be converted to tensors of dtype
+        `tf.float64`.
+
+        :param lower: The lower (inclusive) bounds of the box. Must have shape [D] for positive D,
+            and if a tensor, must have float type.
+        :param upper: The upper (inclusive) bounds of the box. Must have shape [D] for positive D,
+            and if a tensor, must have float type.
+        :raise ValueError (or InvalidArgumentError): If any of the following are true:
+
+            - ``lower`` and ``upper`` have invalid shapes.
+            - ``lower`` and ``upper`` do not have the same floating point type.
+            - ``upper`` is not greater than ``lower`` across all dimensions.
         """
-        :param lower: The lower (inclusive) bounds of the box.
-        :param upper: The upper (inclusive) bounds of the box.
-        :raise ValueError: If ``lower`` and ``upper`` have different shapes. Or if ``upper`` is not
-            greater than ``lower`` across all dimensions.
-        :raise TypeError: If ``lower`` and ``upper`` have different dtypes.
-        """
+        tf.debugging.assert_shapes([(lower, ["D"]), (upper, ["D"])])
+        tf.assert_rank(lower, 1)
+        tf.assert_rank(upper, 1)
 
-        if not shapes_equal(lower, upper):
-            raise ValueError(
-                f"Lower and upper bounds must have the same shape, got {lower.shape} and"
-                f" {upper.shape}"
-            )
+        if len(lower) == 0:
+            raise ValueError(f"Bounds must have shape [D] for positive D, got {tf.shape(lower)}.")
 
-        if lower.dtype is not upper.dtype:
-            raise TypeError(
-                f"Lower and upper bounds must have the same dtype, got {lower.shape} and"
-                f" {upper.shape}"
-            )
+        if isinstance(lower, list):
+            self._lower = tf.cast(lower, dtype=tf.float64)
+            self._upper = tf.cast(upper, dtype=tf.float64)
+        else:
+            self._lower = tf.convert_to_tensor(lower)
+            self._upper = tf.convert_to_tensor(upper)
 
-        if tf.reduce_any(lower >= upper):
-            raise ValueError(
-                f"Upper bound must be greater that lower for all dimensions, got lower bound"
-                f" {lower} and upper {upper}."
-            )
+            tf.debugging.assert_same_float_dtype([self._lower, self._upper])
 
-        self._lower = lower
-        self._upper = upper
+        tf.debugging.assert_less(self._lower, self._upper)
+
+    def __repr__(self) -> str:
+        """"""
+        return f"Box({self._lower!r}, {self._upper!r})"
 
     @property
     def lower(self) -> TensorType:
@@ -191,7 +222,7 @@ class Box(SearchSpace):
         """ The upper bounds of the box. """
         return self._upper
 
-    def __contains__(self, value: TensorType) -> Union[bool, tf.Tensor]:
+    def __contains__(self, value: TensorType) -> Union[bool, TensorType]:
         """
         Return `True` if ``value`` is a member of this search space, else `False`. A point is a
         member if all of its coordinates lie in the closed intervals bounded by the lower and upper
@@ -199,7 +230,7 @@ class Box(SearchSpace):
 
         :param value: A point to check for membership of this :class:`SearchSpace`.
         :return: `True` if ``value`` is a member of this search space, else `False`. May return a
-            scalar boolean `tf.Tensor` instead of the `bool` itself.
+            scalar boolean `TensorType` instead of the `bool` itself.
         :raise ValueError (or InvalidArgumentError): If ``value`` has a different dimensionality
             from the search space.
         """
@@ -211,7 +242,7 @@ class Box(SearchSpace):
 
         return tf.reduce_all(value >= self._lower) and tf.reduce_all(value <= self._upper)
 
-    def sample(self, num_samples: int) -> tf.Tensor:
+    def sample(self, num_samples: int) -> TensorType:
         dim = tf.shape(self._lower)[-1]
         return tf.random.uniform(
             (num_samples, dim), minval=self._lower, maxval=self._upper, dtype=self._lower.dtype
@@ -226,15 +257,30 @@ class Box(SearchSpace):
         return DiscreteSearchSpace(points=self.sample(num_samples))
 
     def __mul__(self, other: Box) -> Box:
-        """
-        Return the Cartesian product of the two :class:`Box`\ es (concatenating their respective lower and upper bounds).
-        :param box: :class:`Box`.
-        :return: the new combined :class:`Box`
-        :raise TypeError: If the lhs and rhs :class:`Box` bounds have different types.
+        r"""
+        Return the Cartesian product of the two :class:`Box`\ es (concatenating their respective
+        lower and upper bounds). For example:
+
+            >>> unit_interval = Box([0.0], [1.0])
+            >>> square_at_origin = Box([-2.0, -2.0], [2.0, 2.0])
+            >>> new_box = unit_interval * square_at_origin
+            >>> new_box.lower.numpy()
+            array([ 0., -2., -2.])
+            >>> new_box.upper.numpy()
+            array([1., 2., 2.])
+
+        :param other: A :class:`Box` with bounds of the same type as this :class:`Box`.
+        :return: The Cartesian product of the two :class:`Box`\ es.
+        :raise TypeError: If the bounds of one :class:`Box` have different dtypes to those of
+            the other :class:`Box`.
         """
         if self.lower.dtype is not other.lower.dtype:
             return NotImplemented
 
-        expanded_lower_bound = tf.concat([self._lower, other.lower], axis=-1)
-        expanded_upper_bound = tf.concat([self._upper, other.upper], axis=-1)
-        return Box(expanded_lower_bound, expanded_upper_bound)
+        product_lower_bound = tf.concat([self._lower, other.lower], axis=-1)
+        product_upper_bound = tf.concat([self._upper, other.upper], axis=-1)
+
+        return Box(product_lower_bound, product_upper_bound)
+
+    def __deepcopy__(self, memo: Dict[int, object]) -> Box:
+        return self
