@@ -12,5 +12,17 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """ This package contains library utilities. """
+from typing_extensions import Final
+
 from . import objectives
-from .misc import jit, shapes_equal, to_numpy
+from .misc import Err, Ok, Result, T_co, jit, shapes_equal, to_numpy
+
+
+class DEFAULTS:
+    """ Default constants used in Trieste. """
+
+    JITTER: Final[float] = 1e-6
+    """
+    The default jitter, typically used to stabilise computations near singular points, such as in
+    Cholesky decomposition.
+    """
